/*
 *  The contents of this file are subject to the Initial
 *  Developer's Public License Version 1.0 (the "License");
 *  you may not use this file except in compliance with the
 *  License. You may obtain a copy of the License at
 *  http://www.ibphoenix.com/main.nfs?a=ibphoenix&page=ibp_idpl.
 *
 *  Software distributed under the License is distributed AS IS,
 *  WITHOUT WARRANTY OF ANY KIND, either express or implied.
 *  See the License for the specific language governing rights
 *  and limitations under the License.
 *
 *  The Original Code was created by Dmitry Yemanov
 *  for the Firebird Open Source RDBMS project.
 *
 *  Copyright (c) 2013 Dmitry Yemanov <dimitr@firebirdsql.org>
 *  and all contributors signed below.
 *
 *  All Rights Reserved.
 *  Contributor(s): ______________________________________.
 */


#ifndef JRD_REPLICATION_H
#define JRD_REPLICATION_H

#include "../../common/classes/timestamp.h"
#include "../../common/os/guid.h"
#include "../../jrd/status.h"

namespace Jrd
{
	class thread_db;
	class jrd_rel;
}

namespace Replication {

	// Supported protocol versions
	const ULONG PROTOCOL_VERSION1 = 1;

	// Protocol packet tags
	const UCHAR eInitialize = 1;
	const UCHAR eCleanupAttachment = 2;
	const UCHAR eStartTransaction = 3;
	const UCHAR ePrepareTransaction = 4;
	const UCHAR eCommitTransaction = 5;
	const UCHAR eRollbackTransaction = 6;
	const UCHAR eStartSavepoint = 7;
	const UCHAR eCleanupSavepoint = 8;
	const UCHAR eInsertRecord = 9;
	const UCHAR eUpdateRecord = 10;
	const UCHAR eDeleteRecord = 11;
	const UCHAR eChangeGenerator = 12;
	const UCHAR eStoreBlob = 13;
	const UCHAR eCleanupTransactions = 14;

	// Option flags
	const ULONG FLAG_COMPRESS_RECORDS	= 1;
	const ULONG FLAG_MASTER_PRIORITY	= 2;
	const ULONG FLAG_IGNORE_MISSING		= 4;
}

namespace Jrd {

class IReplicationMaster
{
public:
	virtual bool cleanupAttachment(FbStatusVector* status, AttNumber att_id) = 0;

	virtual bool startTransaction(FbStatusVector* status, AttNumber att_id, TraNumber txn_id) = 0;
	virtual bool prepareTransaction(FbStatusVector* status, TraNumber txn_id) = 0;
	virtual bool commitTransaction(FbStatusVector* status, TraNumber txn_id) = 0;
	virtual bool rollbackTransaction(FbStatusVector* status, TraNumber txn_id) = 0;
	virtual bool cleanupTransactions(FbStatusVector* status) = 0;

	virtual bool startSavepoint(FbStatusVector* status, TraNumber txn_id, SavNumber sav_id) = 0;
	virtual bool cleanupSavepoint(FbStatusVector* status, TraNumber txn_id, SavNumber sav_id, bool undo) = 0;

	virtual bool insertRecord(FbStatusVector* status, TraNumber txn_id, const char* rel_name,
							  ULONG length, const UCHAR* data) = 0;
	virtual bool updateRecord(FbStatusVector* status, TraNumber txn_id, const char* rel_name,
							  ULONG org_length, const UCHAR* org_data,
							  ULONG new_length, const UCHAR* new_data) = 0;
	virtual bool deleteRecord(FbStatusVector* status, TraNumber txn_id, const char* rel_name,
							  ULONG length, const UCHAR* data) = 0;

	virtual bool changeGenerator(FbStatusVector* status, SLONG gen_id, SINT64 value) = 0;

	virtual bool storeBlob(FbStatusVector* status, TraNumber txn_id, ISC_QUAD blob_id,
						   ULONG length, const UCHAR* data) = 0;

	virtual void shutdown(FbStatusVector* status, bool drop) = 0;

	virtual bool isTableEnabled(Jrd::thread_db* tdbb, Jrd::jrd_rel* const relation) = 0;
	virtual void clearTableEnabled(const char* rel_name) = 0;
	virtual bool isMasterPriority() = 0;
};

class IReplicationHelper
{
public:
	virtual const char* getDatabaseName() = 0;
	virtual const Firebird::Guid* getDatabaseGuid() = 0;
};

void REPL_replicate(thread_db* tdbb, ULONG length, const UCHAR* data);
void REPL_init(thread_db* tdbb, IReplicationHelper* helper,
			   ULONG dpb_length, const UCHAR* dpb, bool create_flag);
void REPL_shutdown(thread_db* tdbb);

} // namespace

#include "../../jrd/repl/Wrappers.h"

#endif // JRD_REPLICATION_H
