/*
 *  The contents of this file are subject to the Initial
 *  Developer's Public License Version 1.0 (the "License");
 *  you may not use this file except in compliance with the
 *  License. You may obtain a copy of the License at
 *  http://www.ibphoenix.com/main.nfs?a=ibphoenix&page=ibp_idpl.
 *
 *  Software distributed under the License is distributed AS IS,
 *  WITHOUT WARRANTY OF ANY KIND, either express or implied.
 *  See the License for the specific language governing rights
 *  and limitations under the License.
 *
 *  The Original Code was created by Dmitry Yemanov
 *  for the Firebird Open Source RDBMS project.
 *
 *  Copyright (c) 2013 Dmitry Yemanov <dimitr@firebirdsql.org>
 *  and all contributors signed below.
 *
 *  All Rights Reserved.
 *  Contributor(s): ______________________________________.
 */

#include "firebird.h"
#include "../jrd/jrd.h"
#include "../jrd/ods.h"
#include "../jrd/req.h"
#include "../jrd/tra.h"
#include "../jrd/blr.h"
#include "../jrd/trig.h"
#include "../jrd/Database.h"
#include "../jrd/blb_proto.h"
#include "../jrd/cch_proto.h"
#include "../jrd/evl_proto.h"
#include "../jrd/met_proto.h"
#include "../jrd/mov_proto.h"

#include "../jrd/repl/Replication.h"

using namespace Firebird;
using namespace Ods;
using namespace Jrd;
using namespace Replication;

namespace
{
	void raiseException(const thread_db* tdbb)
	{
		if (tdbb->tdbb_status_vector->isEmpty())
			LongJump::raise();

		status_exception::raise(tdbb->tdbb_status_vector);
	}

	Record* upgradeRecord(thread_db* tdbb, jrd_rel* relation, Record* record)
	{
		Format* const format = MET_current(tdbb, relation);

		if (record->getFormat()->fmt_version == format->fmt_version)
			return record;

		MemoryPool& pool = *tdbb->getDefaultPool();
		const USHORT length = format->fmt_length;
		Record* const new_record = FB_NEW_POOL(pool) Record(pool, format);

		DSC org_desc, new_desc;

		for (USHORT i = 0; i < new_record->getFormat()->fmt_count; i++)
		{
			new_record->clearNull(i);

			if (EVL_field(relation, new_record, i, &new_desc))
			{
				if (EVL_field(relation, record, i, &org_desc))
					MOV_move(tdbb, &org_desc, &new_desc);
				else
					new_record->setNull(i);
			}
		}

		return new_record;
	}

	void startSavepoints(thread_db* tdbb, Jrd::jrd_rel* relation, jrd_tra* transaction)
	{
		Database* const dbb = tdbb->getDatabase();
		fb_assert(dbb);
		Attachment* const attachment = tdbb->getAttachment();
		fb_assert(attachment);

		IReplicationMaster* const replicator = attachment->att_replicator;
		fb_assert(replicator);

		HalfStaticArray<SLONG, 16> stack;
		for (Savepoint* savepoint = transaction->tra_save_point; savepoint;
			savepoint = savepoint->sav_next)
		{
			if (!(savepoint->sav_flags & SAV_trans_level))
				stack.add(savepoint->sav_number);
		}

		for (Stack<Savepoint*>::iterator iter(transaction->tra_save_proc);
			iter.hasData(); ++iter)
		{
			for (Savepoint* savepoint = iter.object(); savepoint;
				savepoint = savepoint->sav_next)
			{
				if (!(savepoint->sav_flags & SAV_trans_level))
					stack.add(savepoint->sav_number);
			}
		}

		while (stack.hasData())
		{
			if (!replicator->startSavepoint(tdbb->tdbb_status_vector,
											transaction->tra_number,
											stack.pop()))
			{
				raiseException(tdbb);
			}
		}
	}
}


// Replication wrappers

void REPL_detach(thread_db* tdbb)
{
	Database* const dbb = tdbb->getDatabase();
	fb_assert(dbb);
	Attachment* const attachment = tdbb->getAttachment();
	fb_assert(attachment);

	IReplicationMaster* const replicator = attachment->att_replicator;

	if (!replicator)
		return;

	if (dbb->dbb_flags & DBB_bugcheck)
		return;

	bool dropping = false;

	try
	{
		if (!replicator->cleanupAttachment(tdbb->tdbb_status_vector,
										   attachment->att_attachment_id))
		{
			raiseException(tdbb);
		}

		WIN window(HEADER_PAGE_NUMBER);
		Ods::header_page* const header =
			(Ods::header_page*) CCH_FETCH(tdbb, &window, LCK_read, pag_header);
		dropping = (header->hdr_ods_version == 0);
		CCH_RELEASE(tdbb, &window);
	}
	catch (const LongJump&)
	{} // no-op

	FbLocalStatus temp_status;
	replicator->shutdown(&temp_status, dropping);
	attachment->att_replicator = NULL;
}

void REPL_trans_start(thread_db* tdbb, jrd_tra* transaction)
{
	Database* const dbb = tdbb->getDatabase();
	fb_assert(dbb);
	Attachment* const attachment = tdbb->getAttachment();
	fb_assert(attachment);

	IReplicationMaster* const replicator = attachment->att_replicator;

	if (!replicator)
		return;

	if (dbb->dbb_flags & DBB_bugcheck)
		return;

	if (tdbb->tdbb_flags & TDBB_sweeper)
		return;

	if (transaction->tra_flags & (TRA_system | TRA_readonly))
		return;

	try
	{
		if (!replicator->startTransaction(tdbb->tdbb_status_vector,
										  attachment->att_attachment_id,
										  transaction->tra_number))
		{
			raiseException(tdbb);
		}
	}
	catch (const LongJump&)
	{
		FbLocalStatus temp_status;
		replicator->shutdown(&temp_status, false);
		attachment->att_replicator = NULL;
	}
}

void REPL_commit(thread_db* tdbb, jrd_tra* transaction, SLONG new_number)
{
	Database* const dbb = tdbb->getDatabase();
	fb_assert(dbb);
	Attachment* const attachment = tdbb->getAttachment();
	fb_assert(attachment);

	IReplicationMaster* const replicator = attachment->att_replicator;

	if (!replicator)
		return;

	if (dbb->dbb_flags & DBB_bugcheck)
		return;

	if (tdbb->tdbb_flags & TDBB_sweeper)
		return;

	if (transaction->tra_flags & (TRA_system | TRA_readonly))
		return;

	try
	{
		if (!replicator->commitTransaction(tdbb->tdbb_status_vector,
										   transaction->tra_number))
		{
			raiseException(tdbb);
		}

		if (new_number)
		{
			if (!replicator->startTransaction(tdbb->tdbb_status_vector,
											  attachment->att_attachment_id,
											  new_number))
			{
				raiseException(tdbb);
			}
		}
	}
	catch (const LongJump&)
	{
		FbLocalStatus temp_status;
		replicator->shutdown(&temp_status, false);
		attachment->att_replicator = NULL;
	}
}

void REPL_prepare(thread_db* tdbb, jrd_tra* transaction)
{
	Database* const dbb = tdbb->getDatabase();
	fb_assert(dbb);
	Attachment* const attachment = tdbb->getAttachment();
	fb_assert(attachment);

	IReplicationMaster* const replicator = attachment->att_replicator;

	if (!replicator)
		return;

	if (dbb->dbb_flags & DBB_bugcheck)
		return;

	if (tdbb->tdbb_flags & TDBB_sweeper)
		return;

	if (transaction->tra_flags & (TRA_system | TRA_readonly))
		return;

	try
	{
		if (!replicator->prepareTransaction(tdbb->tdbb_status_vector,
											transaction->tra_number))
		{
			raiseException(tdbb);
		}
	}
	catch (const LongJump&)
	{
		FbLocalStatus temp_status;
		replicator->shutdown(&temp_status, false);
		attachment->att_replicator = NULL;
	}
}

void REPL_rollback(thread_db* tdbb, jrd_tra* transaction, SLONG new_number)
{
	Database* const dbb = tdbb->getDatabase();
	fb_assert(dbb);
	Attachment* const attachment = tdbb->getAttachment();
	fb_assert(attachment);

	IReplicationMaster* const replicator = attachment->att_replicator;

	if (!replicator)
		return;

	if (dbb->dbb_flags & DBB_bugcheck)
		return;

	if (tdbb->tdbb_flags & TDBB_sweeper)
		return;

	if (transaction->tra_flags & (TRA_system | TRA_readonly))
		return;

	try
	{
		if (!replicator->rollbackTransaction(tdbb->tdbb_status_vector,
											 transaction->tra_number))
		{
			raiseException(tdbb);
		}

		if (new_number)
		{
			if (!replicator->startTransaction(tdbb->tdbb_status_vector,
											  attachment->att_attachment_id,
											  new_number))
			{
				raiseException(tdbb);
			}
		}
	}
	catch (const LongJump&)
	{
		FbLocalStatus temp_status;
		replicator->shutdown(&temp_status, false);
		attachment->att_replicator = NULL;
	}
}

void REPL_trans_cleanup(Jrd::thread_db* tdbb)
{
	Database* const dbb = tdbb->getDatabase();
	fb_assert(dbb);
	Attachment* const attachment = tdbb->getAttachment();
	fb_assert(attachment);

	IReplicationMaster* const replicator = attachment->att_replicator;

	if (!replicator)
		return;

	if (dbb->dbb_flags & DBB_bugcheck)
		return;

	try
	{
		if (!replicator->cleanupTransactions(tdbb->tdbb_status_vector))
		{
			raiseException(tdbb);
		}
	}
	catch (const LongJump&)
	{
		FbLocalStatus temp_status;
		replicator->shutdown(&temp_status, false);
		attachment->att_replicator = NULL;
	}
}

void REPL_cleanup(thread_db* tdbb, jrd_tra* transaction,
				  const Savepoint* savepoint, bool undo)
{
	Database* const dbb = tdbb->getDatabase();
	fb_assert(dbb);
	Attachment* const attachment = tdbb->getAttachment();
	fb_assert(attachment);

	IReplicationMaster* const replicator = attachment->att_replicator;

	if (!replicator)
		return;

	if (dbb->dbb_flags & DBB_bugcheck)
		return;

	if (tdbb->tdbb_flags & TDBB_dont_post_dfw)
		return;

	if (transaction->tra_flags & (TRA_system | TRA_readonly))
		return;

	if (savepoint->sav_flags & SAV_trans_level)
		return;

	try
	{
		if (!replicator->cleanupSavepoint(tdbb->tdbb_status_vector,
										  transaction->tra_number,
										  savepoint->sav_number,
										  undo))
		{
			raiseException(tdbb);
		}
	}
	catch (const LongJump&)
	{
		FbLocalStatus temp_status;
		replicator->shutdown(&temp_status, false);
		attachment->att_replicator = NULL;
	}
}

void REPL_store(thread_db* tdbb, const record_param* rpb, jrd_tra* transaction)
{
	Database* const dbb = tdbb->getDatabase();
	fb_assert(dbb);
	Attachment* const attachment = tdbb->getAttachment();
	fb_assert(attachment);

	IReplicationMaster* const replicator = attachment->att_replicator;

	if (!replicator)
		return;

	if (dbb->dbb_flags & DBB_bugcheck)
		return;

	if (tdbb->tdbb_flags & TDBB_dont_post_dfw)
		return;

	if (transaction->tra_flags & (TRA_system | TRA_readonly))
		return;

	jrd_rel* const relation = rpb->rpb_relation;
	fb_assert(relation);

	if (relation->isTemporary())
		return;

	if (!relation->isSystem() && !replicator->isTableEnabled(tdbb, relation))
		return;

	try
	{
		Record* const record = upgradeRecord(tdbb, relation, rpb->rpb_record);
		fb_assert(record);

		// The temporary auto-pointer is just to delete a temporary record
		AutoPtr<Record> cleanupRecord(record != rpb->rpb_record ? record : NULL);

		UCharBuffer buffer;
		const Format* const format = record->getFormat();
		for (USHORT id = 0; id < format->fmt_count; id++)
		{
			dsc desc;
			if (DTYPE_IS_BLOB(format->fmt_desc[id].dsc_dtype) &&
				EVL_field(NULL, record, id, &desc))
			{
				const bid* const destination = (bid*) desc.dsc_address;

				if (!destination->isEmpty())
				{
					const ISC_QUAD blob_id =
						{destination->bid_quad.bid_quad_high, destination->bid_quad.bid_quad_low};
					blb* const blob = blb::open(tdbb, transaction, destination);
					UCHAR* const ptr = buffer.getBuffer(blob->blb_length);
					const ULONG len = blob->BLB_get_data(tdbb, ptr, blob->blb_length, true);

					if (!replicator->storeBlob(tdbb->tdbb_status_vector,
											   transaction->tra_number,
											   blob_id, len, ptr))
					{
						raiseException(tdbb);
					}
				}
			}
		}

		startSavepoints(tdbb, relation, transaction);

		if (!replicator->insertRecord(tdbb->tdbb_status_vector,
									  transaction->tra_number,
									  relation->rel_name.c_str(),
									  record->getLength(),
									  record->getData()))
		{
			raiseException(tdbb);
		}
	}
	catch (const LongJump&)
	{
		FbLocalStatus temp_status;
		replicator->shutdown(&temp_status, false);
		attachment->att_replicator = NULL;
	}
}

void REPL_modify(thread_db* tdbb, const record_param* orgRpb,
				 const record_param* newRpb, jrd_tra* transaction)
{
	Database* const dbb = tdbb->getDatabase();
	fb_assert(dbb);
	Attachment* const attachment = tdbb->getAttachment();
	fb_assert(attachment);

	IReplicationMaster* const replicator = attachment->att_replicator;

	if (!replicator)
		return;

	if (dbb->dbb_flags & DBB_bugcheck)
		return;

	if (tdbb->tdbb_flags & TDBB_dont_post_dfw)
		return;

	if (transaction->tra_flags & (TRA_system | TRA_readonly))
		return;

	jrd_rel* const relation = newRpb->rpb_relation;
	fb_assert(relation);

	if (relation->isTemporary())
		return;

	if (!relation->isSystem() && !replicator->isTableEnabled(tdbb, relation))
		return;

	try
	{
		Record* const newRecord = upgradeRecord(tdbb, relation, newRpb->rpb_record);
		fb_assert(newRecord);

		Record* const orgRecord = upgradeRecord(tdbb, relation, orgRpb->rpb_record);
		fb_assert(orgRecord);

		// The temporary auto-pointer is just to delete a temporary record
		AutoPtr<Record> cleanupOrgRecord(orgRecord != orgRpb->rpb_record ? orgRecord : NULL);
		AutoPtr<Record> cleanupNewRecord(newRecord != newRpb->rpb_record ? newRecord : NULL);

		const ULONG org_length = orgRecord->getLength();
		const ULONG new_length = newRecord->getLength();

		// Ignore dummy updates
		if (org_length == new_length &&
			!memcmp(orgRecord->getData(), newRecord->getData(), org_length))
		{
			return;
		}

		const Format* const format = newRecord->getFormat();

		UCharBuffer buffer;
		for (USHORT id = 0; id < format->fmt_count; id++)
		{
			dsc desc;
			if (DTYPE_IS_BLOB(format->fmt_desc[id].dsc_dtype) &&
				EVL_field(NULL, newRecord, id, &desc))
			{
				const bid* const destination = (bid*) desc.dsc_address;

				if (!destination->isEmpty())
				{
					bool store_blob = replicator->isMasterPriority();
					if (!store_blob)
					{
						store_blob = !EVL_field(NULL, orgRecord, id, &desc);
						if (!store_blob)
						{
							const bid* const source = (bid*) desc.dsc_address;
							store_blob = !(*destination == *source);
						}
					}

					if (store_blob)
					{
						const ISC_QUAD blob_id =
							{destination->bid_quad.bid_quad_high, destination->bid_quad.bid_quad_low};
						blb* const blob = blb::open(tdbb, transaction, destination);
						UCHAR* const ptr = buffer.getBuffer(blob->blb_length);
						const ULONG len = blob->BLB_get_data(tdbb, ptr, blob->blb_length, true);

						if (!replicator->storeBlob(tdbb->tdbb_status_vector,
												   transaction->tra_number,
												   blob_id, len, ptr))
						{
							raiseException(tdbb);
						}
					}
				}
			}
		}

		startSavepoints(tdbb, relation, transaction);

		if (!replicator->updateRecord(tdbb->tdbb_status_vector,
									  transaction->tra_number,
									  relation->rel_name.c_str(),
									  org_length,
									  orgRecord->getData(),
									  new_length,
									  newRecord->getData()))
		{
			raiseException(tdbb);
		}
	}
	catch (const LongJump&)
	{
		FbLocalStatus temp_status;
		replicator->shutdown(&temp_status, false);
		attachment->att_replicator = NULL;
	}
}


void REPL_erase(thread_db* tdbb, const record_param* rpb, jrd_tra* transaction)
{
	Database* const dbb = tdbb->getDatabase();
	fb_assert(dbb);
	Attachment* const attachment = tdbb->getAttachment();
	fb_assert(attachment);

	IReplicationMaster* const replicator = attachment->att_replicator;

	if (!replicator)
		return;

	if (dbb->dbb_flags & DBB_bugcheck)
		return;

	if (tdbb->tdbb_flags & TDBB_dont_post_dfw)
		return;

	if (transaction->tra_flags & (TRA_system | TRA_readonly))
		return;

	jrd_rel* const relation = rpb->rpb_relation;
	fb_assert(relation);

	if (relation->isTemporary())
		return;

	if (!relation->isSystem() && !replicator->isTableEnabled(tdbb, relation))
		return;

	try
	{
		Record* const record = upgradeRecord(tdbb, relation, rpb->rpb_record);
		fb_assert(record);

		// The temporary auto-pointer is just to delete a temporary record
		AutoPtr<Record> cleanupRecord(record != rpb->rpb_record ? record : NULL);

		startSavepoints(tdbb, relation, transaction);

		if (!replicator->deleteRecord(tdbb->tdbb_status_vector,
									  transaction->tra_number,
									  relation->rel_name.c_str(),
									  record->getLength(),
									  record->getData()))
		{
			raiseException(tdbb);
		}
	}
	catch (const LongJump&)
	{
		FbLocalStatus temp_status;
		replicator->shutdown(&temp_status, false);
		attachment->att_replicator = NULL;
	}
}

void REPL_gen_id(thread_db* tdbb, SLONG genId, SINT64 value)
{
	Database* const dbb = tdbb->getDatabase();
	fb_assert(dbb);
	Attachment* const attachment = tdbb->getAttachment();
	fb_assert(attachment);

	IReplicationMaster* const replicator = attachment->att_replicator;

	if (!replicator)
		return;

	if (dbb->dbb_flags & DBB_bugcheck)
		return;

	for (const gen* generator = generators; generator->gen_name; generator++)
	{
		if (generator->gen_id == genId)
			return;
	}

	try
	{
		if (!replicator->changeGenerator(tdbb->tdbb_status_vector, genId, value))
		{
			raiseException(tdbb);
		}
	}
	catch (const LongJump&)
	{
		FbLocalStatus temp_status;
		replicator->shutdown(&temp_status, false);
		attachment->att_replicator = NULL;
	}
}
